#include <iostream>
#include <fstream>
#include <stdlib.h>
#include <string>
#include <vector>
#include <pwd.h>
#include <grp.h>
#include <unistd.h>
#include <sys/time.h>
#include <sys/types.h>
#include <rpc/rpc.h>
#include <rpcsvc/ypclnt.h>
// this wasn't needed, and it gives the SGI c++ compiler a bit of a worry
// #include <rpcsvc/yp_prot.h>
#include <crypt.h>

#if defined(__sun) && defined(__SVR4) || defined(__sgi)
	#include <shadow.h>
#endif

#define STRING_LEN 32768

struct passwd *GetPW(void)
	{
	struct passwd *pwent;
	pwent = getpwuid(getuid());
	if (pwent == NULL)
		{
		std::cerr << "You have no password entry.  Sorry." <<
			"  Please tell your sysadmin right away.\n This might be indicative "
			<< "of a serious problem.\n";
		exit(1);
		}
	return pwent;
	}

// splitfields courtesy of "lrobb" <lrobb@home.com> on comp.lang.c++,
// modified slightly by "Dan Stromberg" <strombrg@uci.edu> to include last
// field
std::vector<std::string> splitfields(const std::string &str,char c)
{
    std::vector < std::string > wl;
    int prev (0);

    std::string::size_type pos = str.find (c);
    while (pos != std::string::npos)
    {
        wl.push_back(str.substr(prev,pos-prev));
        prev = pos+1;
        pos = str.find(c,prev);
    }
	 wl.push_back(str.substr(prev,str.length()));
    return wl;
}

// this uses C-style memory allocation, so is a good thing to look at closely
// in an audit
int my_setenv(char *name,char *value)
	{
	char *temp;
	temp = (char *)malloc(strlen(name)+strlen(value) + 2);
	if (temp == NULL)
		{
		std::cerr << "malloc failed in my_setenv\n";
		exit(1);
		}
	strcpy(temp,name);
	strcat(temp,"=");
	strcat(temp,value);
	return putenv(temp);
	}

void give_shell(std::string target_user,char *cmd)
	{
	char *shell;
	struct passwd *pwent;
	pwent = getpwnam(target_user.c_str());
	if (pwent == NULL)
		{
		std::cerr << "User " << target_user << " does not exist\n";
		exit(1);
		}
	shell=pwent->pw_shell;
	if (shell == NULL)
		{
		shell="/bin/sh";
		}
	if (my_setenv("SHELL",shell) != 0)
		{
		std::cerr << "setenv SHELL failed\n";
		exit(1);
		}
	if (my_setenv("HOME",pwent->pw_dir) != 0)
		{
		std::cerr << "setenv HOME failed\n";
		exit(1);
		}
	if (setgid(pwent->pw_gid))
		{
		std::cerr << "setgid failed\n";
		exit(1);
		}
	if (initgroups(pwent->pw_name,pwent->pw_gid) != 0)
		{
		std::cerr << "initgroups failed\n";
		exit(1);
		}
	if (setuid(pwent->pw_uid))
		{
		std::cerr << "setuid failed\n";
		exit(1);
		}
	if (cmd)
		{
		// system seems to lead to hangs when we suspend
		// system(cmd);
		execl(shell,shell,"-c",cmd,NULL);
		perror("execl failed");
		std::cerr << "Exec " << shell << " -c " << cmd << " failed\n";
		}
	else
		{
		execl(shell,shell,NULL);
		perror("execl failed");
		std::cerr << "Exec " << shell << " failed\n";
		}
	exit(1);
	}

char *GetPass(void)
	{
#if defined(__sun) && defined(__SVR4)
	// sun's getpass appears to be bad, but getpassphrase is ok
	return getpassphrase("Password: ");
#elif defined(__linux) || defined(__sgi) || defined(__osf__)
	// getpass is documented as returning a fixed length for these
	return getpass("Password: ");
#else
#	error "You need to define a way of getting a passphrase"
#endif
	}

int correct_passwd(void)
	{
	struct passwd *pwent;
	char *pass;
	char *crypted_text;
	char *real_passwd;
	pwent = GetPW();
#if defined(__sun) && defined(__SVR4) || defined(__sgi)
	if (strlen(pwent->pw_passwd) < 13)
		{
		struct spwd *spwd;
		// cout << "pwent->pw_name is " << pwent->pw_name << endl;
		spwd = getspnam(pwent->pw_name);
		if (spwd == NULL)
			{
			std::cerr << "Could not find shadow password entry\n";
			exit(1);
			}
		real_passwd = spwd->sp_pwdp;
		// cout << "real_passwd is " << real_passwd << endl;
		}
	else
		{
		real_passwd = pwent->pw_passwd;
		}
#else
	real_passwd = pwent->pw_passwd;
#endif
	pass = GetPass();
	// crypt gets key and salt, salt needn't be exactly two chars
	crypted_text = crypt(pass,real_passwd);
	if (crypted_text == NULL)
		{
		std::cerr << "crypt failed\n";
		exit(1);
		}
	if (strcmp(crypted_text,real_passwd) == 0)
		{
		return 1;
		}
	std::cerr << "Sorry.\n";
	exit(0);
	// just to get the compiler to be quiet
	return 0;
	}

const std::string username(void)
	{
	std::string retval;
	struct passwd *pwent;
	pwent = GetPW();
	retval = std::string(pwent->pw_name);
	return retval;
	}

int matched(const std::string target_user,
	const std::string file_target_user,
	const std::string authed_users_string)
	{
	std::string stripped_file_target_user;
	if (file_target_user.at(0) == '~')
		{
		std::string::size_type len;
		/* std::cout << "stripping " << file_target_user << " to "; */
		len = file_target_user.length();
		stripped_file_target_user = file_target_user.substr(1,len);
		/* std::cout << stripped_file_target_user << std::endl; */
		}
	else
		{
		std::cerr << "No ~ on user/group " << file_target_user <<
			"; gsu to group " << "not supported.\n";
		exit(1);
		}
	if (stripped_file_target_user.at(0) == '+')
		{
		std::cerr << "Bogus + on " << stripped_file_target_user << std::endl;;
		exit(1);
		}
	if (target_user == stripped_file_target_user)
		{
		/* cout << "found possible match\n"; */
		std::vector<std::string> authed_users_vector;
		std::string un = username();
		authed_users_vector = splitfields(authed_users_string,',');
		for (std::vector<std::string>::const_iterator
			user=authed_users_vector.begin();
			user != authed_users_vector.end();
			++user)
			{
			/* cout << "checking " << un << " against " << *user << "\n"; */
			if (*user == un)
				{
				/* cout << "matched\n"; */
				return 1;
				}
			}
		return 0;
		}
	else
		{
		return 0;
		}
	return 1;
	}

void usage(char *argv0)
	{
	std::cerr << "Usage: " << argv0 << " username [-c cmd]\n";
	exit(1);
	}

int main(int argc,char *argv[])
	{
	char bad_line[STRING_LEN];
	char *cmd = NULL;
	std::vector < std::string > fields;
	if (argc < 2)
		{
		usage(argv[0]);
		}
	std::string target_user = argv[1];
	if (argc > 2)
		{
		if (argc == 4 && strcmp(argv[2],"-c") == 0)
			{
			cmd = argv[3];
			}
		else
			{
			usage(argv[0]);
			}
		}
	/* cout << "attempting to gsu to " << target_user << std::endl; */
	std::ifstream groupleader_file;
	groupleader_file.open("/var/adm/groupleader",std::ifstream::in);
	if (!groupleader_file.good())
		{
		groupleader_file.open("/usr/local/etc/groupleader",std::ifstream::in);
		if (!groupleader_file.good())
			{
			std::cerr << "No groupleader file found\n";
			exit(1);
			}
		}
#if NEVER
	// simple test of splitfields
	fields = splitfields(":ab:cd::efg:ghi",':');
	// display the fields
	for (std::vector<std::string>::const_iterator field=fields.begin();
		field != fields.end();
		++field)
		{
		cout << *field << ' ' << (*field).length() << ' ';
		}
	cout << std::endl;
#endif
	for (;;)
		{
		groupleader_file.getline(bad_line,STRING_LEN);
		if (!groupleader_file.good())
			break;
		/* cout << "Read groupleader line " << bad_line << std::endl; */
		std::string good_line = bad_line;
		fields = splitfields(good_line,':');
		if (fields[0].at(0) == '+')
			{
			std::string yp_name;
			static char *domainname=NULL;
			int err;
			char *datum;
			int datumlen;
			// yp
			yp_name = fields[0].substr(1,fields[0].length());
			/* cout << "yp line found, yp name is " << yp_name << "\n"; */
			if (domainname == NULL)
				{
				if ((err = yp_get_default_domain(&domainname)) != 0)
					{
					std::cerr << "Error obtaining domainname: " <<
						yperr_string(err) << std::endl;
					exit(1);
					}
				/* cout << "domainname is " << domainname << std::endl; */
				}
			err = yp_match(domainname,"groupleader.bygroup",
				(char *)yp_name.c_str(),(size_t)yp_name.length(),
				&datum,&datumlen);
			if (err)
				{
				std::cerr << "Warning: failed looking up " << yp_name << ": " <<
					yperr_string(err) << std::endl;
				continue;
				}
			fields = splitfields(std::string(datum),':');
			if (fields.size() != 3)
				{
				std::cerr << "Number of colon separated fields must be 3 for:\n";
				std::cerr << datum;
				exit(1);
				}
			if (matched(target_user,fields[0],fields[2]) && correct_passwd())
				{
				give_shell(target_user,cmd);
				}
			}
		else
			{
			/* cout << "nonyp line found\n"; */
			if (fields.size() != 3)
				{
				std::cerr << "wrong number of fields found in nonyp line:\n";
				std::cerr << good_line << std::endl;
				exit(1);
				}
			// users listed in file with unused password field
			if (matched(target_user,fields[0],fields[2]) && correct_passwd())
				{
				give_shell(target_user,cmd);
				}
			}
		}
	groupleader_file.close();
	std::cerr << "Sorry.\n";
	return 0;
	}

